@@ -8,56 +8,39 @@ use Mojo::Home;
 use Mojo::Log;
 use Mojo::Transaction::HTTP;
 use Mojo::UserAgent;
+use Mojo::Util;
 use Scalar::Util 'weaken';
 
 has home => sub { Mojo::Home->new };
 has log  => sub { Mojo::Log->new };
 has ua   => sub {
-  my $self = shift;
-
-  my $ua = Mojo::UserAgent->new->app($self);
-  weaken $self;
-  $ua->on(error => sub { $self->log->error($_[1]) });
-  weaken $ua->{app};
-
+  my $ua = Mojo::UserAgent->new;
+  weaken $ua->server->app(shift)->{app};
   return $ua;
 };
 
+sub build_tx { Mojo::Transaction::HTTP->new }
+
+sub config { Mojo::Util::_stash(config => @_) }
+
+sub handler { croak 'Method "handler" not implemented in subclass' }
+
 sub new {
   my $self = shift->SUPER::new(@_);
 
   # Check if we have a log directory
-  my $home = $self->home->detect(ref $self);
+  my $home = $self->home;
+  $home->detect(ref $self) unless @{$home->parts};
   $self->log->path($home->rel_file('log/mojo.log'))
     if -w $home->rel_file('log');
 
   return $self;
 }
 
-sub build_tx { Mojo::Transaction::HTTP->new }
-
-sub config { shift->_dict(config => @_) }
-
-sub handler { croak 'Method "handler" not implemented in subclass' }
-
-sub _dict {
-  my ($self, $name) = (shift, shift);
-
-  # Hash
-  my $dict = $self->{$name} ||= {};
-  return $dict unless @_;
-
-  # Get
-  return $dict->{$_[0]} unless @_ > 1 || ref $_[0];
-
-  # Set
-  %$dict = (%$dict, %{ref $_[0] ? $_[0] : {@_}});
-
-  return $self;
-}
-
 1;
 
+=encoding utf8
+
 =head1 NAME
 
 Mojo - Duct tape for the HTML5 web!
@@ -87,12 +70,11 @@ Mojo - Duct tape for the HTML5 web!
 
 =head1 DESCRIPTION
 
-Mojo provides a flexible runtime environment for Perl real-time web
-frameworks. It provides all the basic tools and helpers needed to write
-simple web applications and higher level web frameworks, such as
-L<Mojolicious>.
+A flexible runtime environment for Perl real-time web frameworks, with all the
+basic tools and helpers needed to write simple web applications and higher
+level web frameworks, such as L<Mojolicious>.
 
-See L<Mojolicious> for more!
+See L<Mojolicious::Guides> for more!
 
 =head1 ATTRIBUTES
 
@@ -125,25 +107,16 @@ The logging layer of your application, defaults to a L<Mojo::Log> object.
   $app   = $app->ua(Mojo::UserAgent->new);
 
 A full featured HTTP user agent for use in your applications, defaults to a
-L<Mojo::UserAgent> object. Note that this user agent should not be used in
-plugins, since non-blocking requests that are already in progress will
-interfere with new blocking ones.
+L<Mojo::UserAgent> object.
 
   # Perform blocking request
-  my $body = $app->ua->get('mojolicio.us')->res->body;
+  say $app->ua->get('example.com')->res->body;
 
 =head1 METHODS
 
 L<Mojo> inherits all methods from L<Mojo::Base> and implements the following
 new ones.
 
-=head2 new
-
-  my $app = Mojo->new;
-
-Construct a new L<Mojo> application. Will automatically detect your home
-directory and set up logging to C<log/mojo.log> if there's a C<log> directory.
-
 =head2 build_tx
 
   my $tx = $app->build_tx;
@@ -153,17 +126,15 @@ object.
 
 =head2 config
 
-  my $config = $app->config;
-  my $foo    = $app->config('foo');
-  $app       = $app->config({foo => 'bar'});
-  $app       = $app->config(foo => 'bar');
+  my $hash = $app->config;
+  my $foo  = $app->config('foo');
+  $app     = $app->config({foo => 'bar'});
+  $app     = $app->config(foo => 'bar');
 
 Application configuration.
 
-  # Manipulate configuration
-  $app->config->{foo} = 'bar';
-  my $foo = $app->config->{foo};
-  delete $app->config->{foo};
+  # Remove value
+  my $foo = delete $app->config->{foo};
 
 =head2 handler
 
@@ -179,6 +150,14 @@ be overloaded in a subclass.
     ...
   }
 
+=head2 new
+
+  my $app = Mojo->new;
+
+Construct a new L<Mojo> application. Will automatically detect your home
+directory if necessary and set up logging to C<log/mojo.log> if there's a
+C<log> directory.
+
 =head1 SEE ALSO
 
 L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicio.us>.
