@@ -2,57 +2,62 @@ package Mojo::IOLoop::Stream;
 use Mojo::Base 'Mojo::EventEmitter';
 
 use Errno qw(EAGAIN ECONNRESET EINTR EPIPE EWOULDBLOCK);
+use Mojo::IOLoop;
 use Scalar::Util 'weaken';
-use Mojo::Util 'steady_time';
 
-has reactor => sub {
-  require Mojo::IOLoop;
-  Mojo::IOLoop->singleton->reactor;
-};
-has timeout => 15;
+has reactor => sub { Mojo::IOLoop->singleton->reactor };
 
 sub DESTROY { shift->close }
 
-sub new { shift->SUPER::new(handle => shift, buffer => '') }
-
 sub close {
   my $self = shift;
 
-  # Cleanup
-  return unless my $reactor = $self->{reactor};
-  $reactor->remove(delete $self->{timer}) if $self->{timer};
-  return unless my $handle = delete $self->{handle};
+  return unless my $reactor = $self->reactor;
+  return unless my $handle  = delete $self->timeout(0)->{handle};
   $reactor->remove($handle);
-
   close $handle;
-  $self->emit_safe('close');
+  $self->emit('close');
+}
+
+sub close_gracefully {
+  my $self = shift;
+  return $self->{graceful} = 1 if $self->is_writing;
+  $self->close;
 }
 
 sub handle { shift->{handle} }
 
 sub is_readable {
   my $self = shift;
-  $self->{active} = steady_time;
+  $self->_again;
   return $self->{handle} && $self->reactor->is_readable($self->{handle});
 }
 
 sub is_writing {
   my $self = shift;
-  return undef unless exists $self->{handle};
+  return undef unless $self->{handle};
   return !!length($self->{buffer}) || $self->has_subscribers('drain');
 }
 
+sub new { shift->SUPER::new(handle => shift, buffer => '', timeout => 15) }
+
 sub start {
   my $self = shift;
-  return $self->_startup unless $self->{startup}++;
-  return unless delete $self->{paused};
-  $self->reactor->watch($self->{handle}, 1, $self->is_writing);
+
+  # Resume
+  my $reactor = $self->reactor;
+  return $reactor->watch($self->{handle}, 1, $self->is_writing)
+    if delete $self->{paused};
+
+  weaken $self;
+  my $cb = sub { pop() ? $self->_write : $self->_read };
+  $reactor->io($self->timeout($self->{timeout})->{handle} => $cb);
 }
 
 sub stop {
   my $self = shift;
-  return if $self->{paused}++;
-  $self->reactor->watch($self->{handle}, 0, $self->is_writing);
+  $self->reactor->watch($self->{handle}, 0, $self->is_writing)
+    unless $self->{paused}++;
 }
 
 sub steal_handle {
@@ -61,56 +66,54 @@ sub steal_handle {
   return delete $self->{handle};
 }
 
+sub timeout {
+  my $self = shift;
+
+  return $self->{timeout} unless @_;
+
+  my $reactor = $self->reactor;
+  $reactor->remove(delete $self->{timer}) if $self->{timer};
+  return $self unless my $timeout = $self->{timeout} = shift;
+  weaken $self;
+  $self->{timer}
+    = $reactor->timer($timeout => sub { $self->emit('timeout')->close });
+
+  return $self;
+}
+
 sub write {
   my ($self, $chunk, $cb) = @_;
 
   $self->{buffer} .= $chunk;
   if ($cb) { $self->once(drain => $cb) }
-  else     { return $self unless length $self->{buffer} }
+  elsif (!length $self->{buffer}) { return $self }
   $self->reactor->watch($self->{handle}, !$self->{paused}, 1)
     if $self->{handle};
 
   return $self;
 }
 
-sub _read {
+sub _again { $_[0]->reactor->again($_[0]{timer}) if $_[0]{timer} }
+
+sub _error {
   my $self = shift;
 
-  my $read = $self->{handle}->sysread(my $buffer, 131072, 0);
-  unless (defined $read) {
+  # Retry
+  return if $! == EAGAIN || $! == EINTR || $! == EWOULDBLOCK;
 
-    # Retry
-    return if grep { $_ == $! } EAGAIN, EINTR, EWOULDBLOCK;
+  # Closed
+  return $self->close if $! == ECONNRESET || $! == EPIPE;
 
-    # Closed
-    return $self->close if grep { $_ == $! } ECONNRESET, EPIPE;
-
-    # Read error
-    return $self->emit_safe(error => $!)->close;
-  }
-
-  # EOF
-  return $self->close if $read == 0;
-
-  $self->emit_safe(read => $buffer)->{active} = steady_time;
+  # Error
+  $self->emit(error => $!)->close;
 }
 
-sub _startup {
+sub _read {
   my $self = shift;
-
-  # Timeout (ignore 0 timeout)
-  my $reactor = $self->reactor;
-  weaken $self;
-  $self->{timer} = $reactor->recurring(
-    0.5 => sub {
-      return unless my $timeout = $self->timeout;
-      $self->emit_safe('timeout')->close
-        if (steady_time - $self->{active}) >= $timeout;
-    }
-  );
-
-  $self->{active} = steady_time;
-  $reactor->io($self->{handle}, sub { pop() ? $self->_write : $self->_read });
+  my $read = $self->{handle}->sysread(my $buffer, 131072, 0);
+  return $self->_error unless defined $read;
+  return $self->close if $read == 0;
+  $self->emit(read => $buffer)->_again;
 }
 
 sub _write {
@@ -119,29 +122,20 @@ sub _write {
   my $handle = $self->{handle};
   if (length $self->{buffer}) {
     my $written = $handle->syswrite($self->{buffer});
-    unless (defined $written) {
-
-      # Retry
-      return if grep { $_ == $! } EAGAIN, EINTR, EWOULDBLOCK;
-
-      # Closed
-      return $self->close if grep { $_ == $! } ECONNRESET, EPIPE;
-
-      # Write error
-      return $self->emit_safe(error => $!)->close;
-    }
-
-    $self->emit_safe(write => substr($self->{buffer}, 0, $written, ''));
-    $self->{active} = steady_time;
+    return $self->_error unless defined $written;
+    $self->emit(write => substr($self->{buffer}, 0, $written, ''))->_again;
   }
 
-  $self->emit_safe('drain') if !length $self->{buffer};
-  return if $self->is_writing;
-  $self->reactor->watch($handle, !$self->{paused}, 0);
+  $self->emit('drain') if !length $self->{buffer};
+  return               if $self->is_writing;
+  return $self->close  if $self->{graceful};
+  $self->reactor->watch($handle, !$self->{paused}, 0) if $self->{handle};
 }
 
 1;
 
+=encoding utf8
+
 =head1 NAME
 
 Mojo::IOLoop::Stream - Non-blocking I/O stream
@@ -189,7 +183,7 @@ emit the following new ones.
     ...
   });
 
-Emitted safely if the stream gets closed.
+Emitted if the stream gets closed.
 
 =head2 drain
 
@@ -198,7 +192,7 @@ Emitted safely if the stream gets closed.
     ...
   });
 
-Emitted safely once all data has been written.
+Emitted once all data has been written.
 
 =head2 error
 
@@ -207,7 +201,7 @@ Emitted safely once all data has been written.
     ...
   });
 
-Emitted safely if an error occurs on the stream.
+Emitted if an error occurs on the stream, fatal if unhandled.
 
 =head2 read
 
@@ -216,7 +210,7 @@ Emitted safely if an error occurs on the stream.
     ...
   });
 
-Emitted safely if new data arrives on the stream.
+Emitted if new data arrives on the stream.
 
 =head2 timeout
 
@@ -225,8 +219,8 @@ Emitted safely if new data arrives on the stream.
     ...
   });
 
-Emitted safely if the stream has been inactive for too long and will get
-closed automatically.
+Emitted if the stream has been inactive for too long and will get closed
+automatically.
 
 =head2 write
 
@@ -235,7 +229,7 @@ closed automatically.
     ...
   });
 
-Emitted safely if new data has been written to the stream.
+Emitted if new data has been written to the stream.
 
 =head1 ATTRIBUTES
 
@@ -246,35 +240,26 @@ L<Mojo::IOLoop::Stream> implements the following attributes.
   my $reactor = $stream->reactor;
   $stream     = $stream->reactor(Mojo::Reactor::Poll->new);
 
-Low level event reactor, defaults to the C<reactor> attribute value of the
+Low-level event reactor, defaults to the C<reactor> attribute value of the
 global L<Mojo::IOLoop> singleton.
 
-=head2 timeout
-
-  my $timeout = $stream->timeout;
-  $stream     = $stream->timeout(45);
-
-Maximum amount of time in seconds stream can be inactive before getting closed
-automatically, defaults to C<15>. Setting the value to C<0> will allow this
-stream to be inactive indefinitely.
-
 =head1 METHODS
 
 L<Mojo::IOLoop::Stream> inherits all methods from L<Mojo::EventEmitter> and
 implements the following new ones.
 
-=head2 new
-
-  my $stream = Mojo::IOLoop::Stream->new($handle);
-
-Construct a new L<Mojo::IOLoop::Stream> object.
-
 =head2 close
 
   $stream->close;
 
 Close stream immediately.
 
+=head2 close_gracefully
+
+  $stream->close_gracefully;
+
+Close stream gracefully.
+
 =head2 handle
 
   my $handle = $stream->handle;
@@ -283,17 +268,23 @@ Get handle for stream.
 
 =head2 is_readable
 
-  my $success = $stream->is_readable;
+  my $bool = $stream->is_readable;
 
 Quick non-blocking check if stream is readable, useful for identifying tainted
 sockets.
 
 =head2 is_writing
 
-  my $success = $stream->is_writing;
+  my $bool = $stream->is_writing;
 
 Check if stream is writing.
 
+=head2 new
+
+  my $stream = Mojo::IOLoop::Stream->new($handle);
+
+Construct a new L<Mojo::IOLoop::Stream> object.
+
 =head2 start
 
   $stream->start;
@@ -312,6 +303,15 @@ Stop watching for new data on the stream.
 
 Steal handle from stream and prevent it from getting closed automatically.
 
+=head2 timeout
+
+  my $timeout = $stream->timeout;
+  $stream     = $stream->timeout(45);
+
+Maximum amount of time in seconds stream can be inactive before getting closed
+automatically, defaults to C<15>. Setting the value to C<0> will allow this
+stream to be inactive indefinitely.
+
 =head2 write
 
   $stream = $stream->write($bytes);
