@@ -6,21 +6,15 @@ use Mojo::Asset::File;
 use Mojo::Asset::Memory;
 use Mojo::Content::MultiPart;
 use Mojo::Content::Single;
-use Mojo::JSON;
+use Mojo::JSON 'encode_json';
 use Mojo::Parameters;
 use Mojo::Transaction::HTTP;
 use Mojo::Transaction::WebSocket;
 use Mojo::URL;
-use Mojo::Util qw(deprecated encode);
+use Mojo::Util qw(encode url_escape);
 
-has generators => sub { {} };
-
-sub new {
-  my $self = shift->SUPER::new(@_);
-  $self->add_generator(form => \&_form);
-  $self->add_generator(json => \&_json);
-  return $self;
-}
+has generators => sub { {form => \&_form, json => \&_json} };
+has name => 'Mojolicious (Perl)';
 
 sub add_generator {
   my ($self, $name, $cb) = @_;
@@ -39,34 +33,15 @@ sub endpoint {
   my $port  = $url->port || ($proto eq 'https' ? 443 : 80);
 
   # Proxy for normal HTTP requests
+  my $socks;
+  if (my $proxy = $req->proxy) { $socks = $proxy->protocol eq 'socks' }
   return $self->_proxy($tx, $proto, $host, $port)
-    if $proto eq 'http' && lc($req->headers->upgrade // '') ne 'websocket';
+    if $proto eq 'http' && !$req->is_handshake && !$socks;
 
   return $proto, $host, $port;
 }
 
-# DEPRECATED in Rainbow!
-sub form {
-  deprecated 'Mojo::UserAgent::Transactor::form is DEPRECATED in favor of '
-    . 'Mojo::UserAgent::Transactor::tx';
-  my ($self, $url, $charset) = (shift, shift, shift);
-  my $form = ref $charset ? $charset : shift;
-  $charset = undef if ref $charset;
-  return $self->tx(POST => $url, @_, form => $form, charset => $charset);
-}
-
-# DEPRECATED in Rainbow!
-sub json {
-  deprecated 'Mojo::UserAgent::Transactor::json is DEPRECATED in favor of '
-    . 'Mojo::UserAgent::Transactor::tx';
-  my ($self, $url, $data) = (shift, shift, shift);
-  return $self->tx(POST => $url, @_, json => $data);
-}
-
-sub peer {
-  my ($self, $tx) = @_;
-  return $self->_proxy($tx, $self->endpoint($tx));
-}
+sub peer { $_[0]->_proxy($_[1], $_[0]->endpoint($_[1])) }
 
 sub proxy_connect {
   my ($self, $old) = @_;
@@ -77,15 +52,16 @@ sub proxy_connect {
 
   # No proxy
   return undef unless my $proxy = $req->proxy;
+  return undef if $proxy->protocol eq 'socks';
 
   # WebSocket and/or HTTPS
   my $url = $req->url;
-  my $upgrade = lc($req->headers->upgrade // '');
-  return undef unless $upgrade eq 'websocket' || $url->protocol eq 'https';
+  return undef unless $req->is_handshake || $url->protocol eq 'https';
 
-  # CONNECT request
+  # CONNECT request (expect a bad response)
   my $new = $self->tx(CONNECT => $url->clone->userinfo(undef));
   $new->req->proxy($proxy);
+  $new->res->content->auto_relax(0)->headers->connection('keep-alive');
 
   return $new;
 }
@@ -95,25 +71,31 @@ sub redirect {
 
   # Commonly used codes
   my $res = $old->res;
-  my $code = $res->code // '';
-  return undef unless grep { $_ eq $code } 301, 302, 303, 307, 308;
+  my $code = $res->code // 0;
+  return undef unless grep { $_ == $code } 301, 302, 303, 307, 308;
 
-  # Fix broken location without authority and/or scheme
-  return unless my $location = $res->headers->location;
+  # Fix location without authority and/or scheme
+  return undef unless my $location = $res->headers->location;
   $location = Mojo::URL->new($location);
   $location = $location->base($old->req->url)->to_abs unless $location->is_abs;
+  my $proto = $location->protocol;
+  return undef unless $proto eq 'http' || $proto eq 'https';
 
   # Clone request if necessary
-  my $new    = Mojo::Transaction::HTTP->new;
-  my $req    = $old->req;
-  my $method = uc $req->method;
-  if (grep { $_ eq $code } 301, 307, 308) {
-    return undef unless my $req = $req->clone;
-    $new->req($req);
-    $req->headers->remove('Host')->remove('Cookie')->remove('Referer');
+  my $new = Mojo::Transaction::HTTP->new;
+  my $req = $old->req;
+  if ($code == 307 || $code == 308) {
+    return undef unless my $clone = $req->clone;
+    $new->req($clone);
   }
-  elsif ($method ne 'HEAD') { $method = 'GET' }
-  $new->req->method($method)->url($location);
+  else {
+    my $method = uc $req->method;
+    my $headers = $new->req->method($method eq 'POST' ? 'GET' : $method)
+      ->content->headers($req->headers->clone)->headers;
+    $headers->remove($_) for grep {/^content-/i} @{$headers->names};
+  }
+  my $headers = $new->req->url($location)->headers;
+  $headers->remove($_) for qw(Authorization Cookie Host Referer);
   return $new->previous($old);
 }
 
@@ -127,8 +109,11 @@ sub tx {
   $url = "http://$url" unless $url =~ m!^/|://!;
   ref $url ? $req->url($url) : $req->url->parse($url);
 
-  # Headers
-  $req->headers->from_hash(shift) if ref $_[0] eq 'HASH';
+  # Headers (we identify ourselves and accept gzip compression)
+  my $headers = $req->headers;
+  $headers->from_hash(shift) if ref $_[0] eq 'HASH';
+  $headers->user_agent($self->name) unless $headers->user_agent;
+  $headers->accept_encoding('gzip') unless $headers->accept_encoding;
 
   # Generator
   if (@_ > 1) {
@@ -144,8 +129,8 @@ sub tx {
 
 sub upgrade {
   my ($self, $tx) = @_;
-  my $code = $tx->res->code // '';
-  return undef unless $tx->req->headers->upgrade && $code eq '101';
+  my $code = $tx->res->code // 0;
+  return undef unless $tx->req->is_handshake && $code == 101;
   my $ws = Mojo::Transaction::WebSocket->new(handshake => $tx, masked => 1);
   return $ws->client_challenge ? $ws : undef;
 }
@@ -154,8 +139,11 @@ sub websocket {
   my $self = shift;
 
   # New WebSocket transaction
-  my $tx    = $self->tx(GET => @_);
-  my $url   = $tx->req->url;
+  my $sub = ref $_[-1] eq 'ARRAY' ? pop : [];
+  my $tx = $self->tx(GET => @_);
+  my $req = $tx->req;
+  $req->headers->sec_websocket_protocol(join ', ', @$sub) if @$sub;
+  my $url   = $req->url;
   my $proto = $url->protocol;
   $url->scheme($proto eq 'wss' ? 'https' : 'http') if $proto;
 
@@ -169,19 +157,19 @@ sub _form {
   my ($self, $tx, $form, %options) = @_;
 
   # Check for uploads and force multipart if necessary
-  my $multipart;
+  my $req       = $tx->req;
+  my $headers   = $req->headers;
+  my $multipart = ($headers->content_type // '') =~ m!multipart/form-data!i;
   for my $value (map { ref $_ eq 'ARRAY' ? @$_ : $_ } values %$form) {
     ++$multipart and last if ref $value eq 'HASH';
   }
-  my $req     = $tx->req;
-  my $headers = $req->headers;
-  $headers->content_type('multipart/form-data') if $multipart;
 
   # Multipart
-  if (($headers->content_type // '') eq 'multipart/form-data') {
+  if ($multipart) {
     my $parts = $self->_multipart($options{charset}, $form);
     $req->content(
       Mojo::Content::MultiPart->new(headers => $headers, parts => $parts));
+    _type($headers, 'multipart/form-data');
     return $tx;
   }
 
@@ -192,16 +180,14 @@ sub _form {
   if ($method eq 'GET' || $method eq 'HEAD') { $req->url->query->merge($p) }
   else {
     $req->body($p->to_string);
-    $headers->content_type('application/x-www-form-urlencoded');
+    _type($headers, 'application/x-www-form-urlencoded');
   }
   return $tx;
 }
 
 sub _json {
   my ($self, $tx, $data) = @_;
-  $tx->req->body(Mojo::JSON->new->encode($data));
-  my $headers = $tx->req->headers;
-  $headers->content_type('application/json') unless $headers->content_type;
+  _type($tx->req->body(encode_json $data)->headers, 'application/json');
   return $tx;
 }
 
@@ -223,7 +209,7 @@ sub _multipart {
         if (my $file = delete $value->{file}) {
           $file = Mojo::Asset::File->new(path => $file) unless ref $file;
           $part->asset($file);
-          $value->{filename} ||= basename $file->path
+          $value->{filename} //= basename $file->path
             if $file->isa('Mojo::Asset::File');
         }
 
@@ -233,7 +219,7 @@ sub _multipart {
         }
 
         # Filename and headers
-        $filename = delete $value->{filename} || $name;
+        $filename = url_escape delete $value->{filename} // $name, '"';
         $filename = encode $charset, $filename if $charset;
         $headers->from_hash($value);
       }
@@ -245,9 +231,10 @@ sub _multipart {
       }
 
       # Content-Disposition
+      $name = url_escape $name, '"';
       $name = encode $charset, $name if $charset;
       my $disposition = qq{form-data; name="$name"};
-      $disposition .= qq{; filename="$filename"} if $filename;
+      $disposition .= qq{; filename="$filename"} if defined $filename;
       $headers->content_disposition($disposition);
     }
   }
@@ -268,8 +255,12 @@ sub _proxy {
   return $proto, $host, $port;
 }
 
+sub _type { $_[0]->content_type($_[1]) unless $_[0]->content_type }
+
 1;
 
+=encoding utf8
+
 =head1 NAME
 
 Mojo::UserAgent::Transactor - User agent transactor
@@ -280,12 +271,12 @@ Mojo::UserAgent::Transactor - User agent transactor
 
   # Simple GET request
   my $t = Mojo::UserAgent::Transactor->new;
-  say $t->tx(GET => 'http://mojolicio.us')->req->to_string;
+  say $t->tx(GET => 'http://example.com')->req->to_string;
 
   # PATCH request with "Do Not Track" header and content
-  say $t->tx(PATCH => 'mojolicio.us' => {DNT => 1} => 'Hi!')->req->to_string;
+  say $t->tx(PATCH => 'example.com' => {DNT => 1} => 'Hi!')->req->to_string;
 
-  # POST request with form data
+  # POST request with form-data
   say $t->tx(POST => 'example.com' => form => {a => 'b'})->req->to_string;
 
   # PUT request with JSON data
@@ -296,6 +287,23 @@ Mojo::UserAgent::Transactor - User agent transactor
 L<Mojo::UserAgent::Transactor> is the transaction building and manipulation
 framework used by L<Mojo::UserAgent>.
 
+=head1 GENERATORS
+
+These content generators are available by default.
+
+=head2 form
+
+  $t->tx(POST => 'http://example.com' => form => {a => 'b'});
+
+Generate query string, C<application/x-www-form-urlencoded> or
+C<multipart/form-data> content.
+
+=head2 json
+
+  $t->tx(PATCH => 'http://example.com' => json => {a => 'b'});
+
+Generate JSON content with L<Mojo::JSON>.
+
 =head1 ATTRIBUTES
 
 L<Mojo::UserAgent::Transactor> implements the following attributes.
@@ -305,20 +313,22 @@ L<Mojo::UserAgent::Transactor> implements the following attributes.
   my $generators = $t->generators;
   $t             = $t->generators({foo => sub {...}});
 
-Registered content generators.
+Registered content generators, by default only C<form> and C<json> are already
+defined.
+
+=head2 name
+
+  my $name = $t->name;
+  $t       = $t->name('Mojolicious');
+
+Value for C<User-Agent> request header of generated transactions, defaults to
+C<Mojolicious (Perl)>.
 
 =head1 METHODS
 
 L<Mojo::UserAgent::Transactor> inherits all methods from L<Mojo::Base> and
 implements the following new ones.
 
-=head2 new
-
-  my $t = Mojo::UserAgent::Transactor->new;
-
-Construct a new transactor and register C<form> and C<json> content
-generators.
-
 =head2 add_generator
 
   $t = $t->add_generator(foo => sub {...});
@@ -341,7 +351,7 @@ Actual peer for transaction.
 
   my $tx = $t->proxy_connect(Mojo::Transaction::HTTP->new);
 
-Build L<Mojo::Transaction::HTTP> proxy connect request for transaction if
+Build L<Mojo::Transaction::HTTP> proxy C<CONNECT> request for transaction if
 possible.
 
 =head2 redirect
@@ -355,66 +365,109 @@ C<307> or C<308> redirect response if possible.
 
   my $tx = $t->tx(GET  => 'example.com');
   my $tx = $t->tx(POST => 'http://example.com');
-  my $tx = $t->tx(GET  => 'http://example.com' => {DNT => 1});
+  my $tx = $t->tx(GET  => 'http://example.com' => {Accept => '*/*'});
   my $tx = $t->tx(PUT  => 'http://example.com' => 'Hi!');
   my $tx = $t->tx(PUT  => 'http://example.com' => form => {a => 'b'});
   my $tx = $t->tx(PUT  => 'http://example.com' => json => {a => 'b'});
-  my $tx = $t->tx(POST => 'http://example.com' => {DNT => 1} => 'Hi!');
+  my $tx = $t->tx(POST => 'http://example.com' => {Accept => '*/*'} => 'Hi!');
   my $tx = $t->tx(
-    PUT  => 'http://example.com' => {DNT => 1} => form => {a => 'b'});
+    PUT  => 'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $t->tx(
-    PUT  => 'http://example.com' => {DNT => 1} => json => {a => 'b'});
+    PUT  => 'http://example.com' => {Accept => '*/*'} => json => {a => 'b'});
 
 Versatile general purpose L<Mojo::Transaction::HTTP> transaction builder for
-requests, with support for content generators.
+requests, with support for L</"GENERATORS">.
 
-  # Inspect generated request
-  say $t->tx(GET => 'mojolicio.us' => {DNT => 1} => 'Bye!')->req->to_string;
+  # Generate and inspect custom GET request with DNT header and content
+  say $t->tx(GET => 'example.com' => {DNT => 1} => 'Bye!')->req->to_string;
 
-  # Streaming response
-  my $tx = $t->tx(GET => 'http://mojolicio.us');
-  $tx->res->body(sub { say $_[1] });
-
-  # Custom socket
-  my $tx = $t->tx(GET => 'http://mojolicio.us');
+  # Use a custom socket for processing this transaction
+  my $tx = $t->tx(GET => 'http://example.com');
   $tx->connection($sock);
 
-  # Generate query parameters
+  # Stream response content to STDOUT
+  my $tx = $t->tx(GET => 'http://example.com');
+  $tx->res->content->unsubscribe('read')->on(read => sub { say $_[1] });
+
+  # PUT request with content streamed from file
+  my $tx = $t->tx(PUT => 'http://example.com');
+  $tx->req->content->asset(Mojo::Asset::File->new(path => '/foo.txt'));
+
+The C<json> content generator uses L<Mojo::JSON> for encoding and sets the
+content type to C<application/json>.
+
+  # POST request with "application/json" content
+  my $tx = $t->tx(
+    POST => 'http://example.com' => json => {a => 'b', c => [1, 2, 3]});
+
+The C<form> content generator will automatically use query parameters for
+C<GET> and C<HEAD> requests.
+
+  # GET request with query parameters
   my $tx = $t->tx(GET => 'http://example.com' => form => {a => 'b'});
 
-  # Use form generator with custom charset
+For all other request methods the C<application/x-www-form-urlencoded> content
+type is used.
+
+  # POST request with "application/x-www-form-urlencoded" content
   my $tx = $t->tx(
-    PUT => 'http://example.com' => form => {a => 'b'} => charset => 'UTF-8');
+    POST => 'http://example.com' => form => {a => 'b', c => 'd'});
 
-  # Multiple form values with the same name
-  my $tx = $t->tx(PUT => 'http://example.com' => form => {a => [qw(b c d)]});
+Parameters may be encoded with the C<charset> option.
 
-  # Multipart upload streamed from file
+  # PUT request with Shift_JIS encoded form values
   my $tx = $t->tx(
-    PUT => 'http://example.com' => form => {mytext => {file => '/foo.txt'}});
+    PUT => 'example.com' => form => {a => 'b'} => charset => 'Shift_JIS');
 
-  # Multipart upload with in-memory content
+An array reference can be used for multiple form values sharing the same name.
+
+  # POST request with form values sharing the same name
+  my $tx = $t->tx(POST => 'http://example.com' => form => {a => [qw(b c d)]});
+
+A hash reference with a C<content> or C<file> value can be used to switch to
+the C<multipart/form-data> content type for file uploads.
+
+  # POST request with "multipart/form-data" content
   my $tx = $t->tx(
     POST => 'http://example.com' => form => {mytext => {content => 'lala'}});
 
-  # Upload multiple files
+  # POST request with multiple files sharing the same name
   my $tx = $t->tx(POST => 'http://example.com' =>
     form => {mytext => [{content => 'first'}, {content => 'second'}]});
 
-  # Customized upload with filename and header
+The C<file> value should contain the path to the file you want to upload or an
+asset object, like L<Mojo::Asset::File> or L<Mojo::Asset::Memory>.
+
+  # POST request with upload streamed from file
+  my $tx = $t->tx(
+    POST => 'http://example.com' => form => {mytext => {file => '/foo.txt'}});
+
+  # POST request with upload streamed from asset
+  my $asset = Mojo::Asset::Memory->new->add_chunk('lalala');
+  my $tx    = $t->tx(
+    POST => 'http://example.com' => form => {mytext => {file => $asset}});
+
+A C<filename> value will be generated automatically, but can also be set
+manually if necessary. All remainging values in the hash reference get merged
+into the C<multipart/form-data> content as headers.
+
+  # POST request with form values and customized upload (filename and header)
   my $tx = $t->tx(POST => 'http://example.com' => form => {
-    myzip => {
-      file     => Mojo::Asset::Memory->new->add_chunk('lalala'),
-      filename => 'foo.zip',
-      DNT      => 1
+    a      => 'b',
+    c      => 'd',
+    mytext => {
+      content        => 'lalala',
+      filename       => 'foo.txt',
+      'Content-Type' => 'text/plain'
     }
   });
 
-The C<form> content generator will automatically use query parameters for
-C<GET>/C<HEAD> requests and the "application/x-www-form-urlencoded" content
-type for everything else. Both get upgraded automatically to using the
-"multipart/form-data" content type when necessary or when the header has been
-set manually.
+The C<multipart/form-data> content type can also be enforced by setting the
+C<Content-Type> header manually.
+
+  # Force "multipart/form-data"
+  my $headers = {'Content-Type' => 'multipart/form-data'};
+  my $tx = $t->tx(POST => 'example.com' => $headers => form => {a => 'b'});
 
 =head2 upgrade
 
@@ -425,8 +478,8 @@ handshake if possible.
 
 =head2 websocket
 
-  my $tx = $t->websocket('ws://localhost:3000');
-  my $tx = $t->websocket('ws://localhost:3000' => {DNT => 1});
+  my $tx = $t->websocket('ws://example.com');
+  my $tx = $t->websocket('ws://example.com' => {DNT => 1} => ['v1.proto']);
 
 Versatile L<Mojo::Transaction::HTTP> transaction builder for WebSocket
 handshake requests.
