@@ -8,13 +8,16 @@ has [qw(cookie_domain secure)];
 has cookie_name        => 'mojolicious';
 has cookie_path        => '/';
 has default_expiration => 3600;
+has deserialize        => sub { \&Mojo::JSON::j };
+has serialize          => sub { \&Mojo::JSON::encode_json };
 
 sub load {
   my ($self, $c) = @_;
 
+  return unless $c->req->headers->cookie;
   return unless my $value = $c->signed_cookie($self->cookie_name);
-  $value =~ s/-/=/g;
-  return unless my $session = Mojo::JSON->new->decode(b64_decode $value);
+  $value =~ y/-/=/;
+  return unless my $session = $self->deserialize->(b64_decode $value);
 
   # "expiration" value is inherited
   my $expiration = $session->{expiration} // $self->default_expiration;
@@ -37,8 +40,7 @@ sub store {
 
   # Don't reset flash for static files
   my $old = delete $session->{flash};
-  @{$session->{new_flash}}{keys %$old} = values %$old
-    if $stash->{'mojo.static'};
+  $session->{new_flash} = $old if $stash->{'mojo.static'};
   delete $session->{new_flash} unless keys %{$session->{new_flash}};
 
   # Generate "expires" value from "expiration" if necessary
@@ -47,8 +49,8 @@ sub store {
   $session->{expires} = $default || time + $expiration
     if $expiration || $default;
 
-  my $value = b64_encode(Mojo::JSON->new->encode($session), '');
-  $value =~ s/=/-/g;
+  my $value = b64_encode($self->serialize->($session), '');
+  $value =~ y/=/-/;
   my $options = {
     domain   => $self->cookie_domain,
     expires  => $session->{expires},
@@ -61,6 +63,8 @@ sub store {
 
 1;
 
+=encoding utf8
+
 =head1 NAME
 
 Mojolicious::Sessions - Signed cookie based session manager
@@ -76,9 +80,9 @@ Mojolicious::Sessions - Signed cookie based session manager
 =head1 DESCRIPTION
 
 L<Mojolicious::Sessions> manages simple signed cookie based sessions for
-L<Mojolicious>. All data gets serialized with L<Mojo::JSON> and stored
-C<Base64> encoded on the client-side, but is protected from unwanted changes
-with a C<HMAC-SHA1> signature.
+L<Mojolicious>. All data gets serialized with L<Mojo::JSON> and stored Base64
+encoded on the client-side, but is protected from unwanted changes with a
+HMAC-SHA1 signature.
 
 =head1 ATTRIBUTES
 
@@ -125,18 +129,43 @@ C<expiration> and C<expires> session values.
   # Delete whole session by setting an expiration date in the past
   $c->session(expires => 1);
 
+=head2 deserialize
+
+  my $cb    = $sessions->deserialize;
+  $sessions = $sessions->deserialize(sub {...});
+
+A callback used to deserialize sessions, defaults to L<Mojo::JSON/"j">.
+
+  $sessions->deserialize(sub {
+    my $bytes = shift;
+    return {};
+  });
+
 =head2 secure
 
-  my $secure = $sessions->secure;
-  $sessions  = $sessions->secure(1);
+  my $bool  = $sessions->secure;
+  $sessions = $sessions->secure($bool);
 
 Set the secure flag on all session cookies, so that browsers send them only
 over HTTPS connections.
 
+=head2 serialize
+
+  my $cb    = $sessions->serialize;
+  $sessions = $sessions->serialize(sub {...});
+
+A callback used to serialize sessions, defaults to
+L<Mojo::JSON/"encode_json">.
+
+  $sessions->serialize(sub {
+    my $hash = shift;
+    return '';
+  });
+
 =head1 METHODS
 
 L<Mojolicious::Sessions> inherits all methods from L<Mojo::Base> and
-implements the following ones.
+implements the following new ones.
 
 =head2 load
 
